#include "MainDefWindow.h"

MainDefWindow::MainDefWindow()
{
	// Nastaveni titulku ona.
	setWindowTitle("Scenario_definition");

	// Inicializace promennych.
	ageValue = 0;
	diabetValue = false;
	numOfDays = 0;
	numOfWeeks = 0;
	currentDay = 0;
	currentWeek = 0;
	progressValue = 0;
	progressIncrease = 1;
	//

	// Naplneni vektrou widgetu (jeden den).
	eventsWidgets.push_back(new DayEventWidgets("Breakfast"));
	eventsWidgets.push_back(new DayEventWidgets("Morning snack"));
	eventsWidgets.push_back(new DayEventWidgets("Lunch"));
	eventsWidgets.push_back(new DayEventWidgets("Afternoon snack"));
	eventsWidgets.push_back(new DayEventWidgets("Dinner"));
	eventsWidgets.push_back(new DayEventWidgets("Movement"));
	//

	// Jeden krok fyzicke aktivity.
	eventsWidgets.back()->getEventBox()->setSingleStep(SINGLE_STEP);

	// Maximalni hodnota fyzocke aktivity.
	eventsWidgets.back()->getEventBox()->setMaximum(MAX_STEP);

	// Inicializace widgetu.
	centralWidget = new QWidget();
	stackedWidget = new QStackedWidget();
	centralLayout = new QGridLayout;

	//dayFrame = new QFrame();

	progress = new QProgressBar();
	fileNAmeEdit = new QLineEdit("generated_scenario.csv");
	prevButton = new QPushButton("<");
	nextButton = new QPushButton(">");
	startButton = new QPushButton("Start");
	statusLabel = new QLabel("Simulation not running");
}


MainDefWindow::~MainDefWindow()
{
	// Na konci chceme, aby se korektne simulace ukoncila.
	if (smartCGMSExecutionThread && filterExecutor)
	{
		// Pokud je potreba posle se signal.
		if (filterExecutor)
			Send_Shut_Down();

		// Veschny vlakna se preponou do join stavu.
		if (smartCGMSExecutionThread->joinable())
			smartCGMSExecutionThread->join();

		// Reser pointeru.
		filterExecutor.reset();
		smartCGMSExecutionThread.reset();
	}

	// Vycisteni vektru.
	eventsWidgets.clear();
}

void MainDefWindow::initWindow()
{

	for (int i = 0; i < numOfWeeks; ++i) {
		// Pridani tyden do vektoru.
		weeks.push_back(new Week(numOfDays, i + 1));
		for (auto& day : weeks[i]->getDays()) {
			// Pridani dne do widgetu
			stackedWidget->addWidget(day->getDayFrame());
		}
	}

	int x = 0;

	// Nastaveni widgetu.
	centralLayout->addWidget(stackedWidget, 0, 1, 1, 1);
	centralWidget->setLayout(centralLayout);
	setCentralWidget(centralWidget);

	progressValue = progressIncrease;
	progress->setMaximum(weeks.size() * weeks[0]->getDays().size());
	progress->setValue(progressValue);
	progress->setTextVisible(false);
	centralLayout->addWidget(progress, 1, 1, 1, 1);

	centralLayout->addWidget(fileNAmeEdit, 2, 1, 1, 1);

	prevButton->setSizePolicy(QSizePolicy::Expanding, QSizePolicy::Expanding);
	prevButton->setMaximumHeight(200);
	prevButton->setMaximumWidth(20);
	prevButton->setDisabled(true);
	centralLayout->addWidget(prevButton, 0, 0, 1, 1, Qt::AlignRight);

	nextButton->setSizePolicy(QSizePolicy::Expanding, QSizePolicy::Expanding);
	nextButton->setMaximumHeight(200);
	nextButton->setMaximumWidth(20);
	if ((weeks.size() * weeks[0]->getDays().size()) == 1) {
		nextButton->setDisabled(true);
	}

	centralLayout->addWidget(nextButton, 0, 2, 1, 1, Qt::AlignLeft);
	centralLayout->addWidget(startButton, 4, 0, 1, 3, Qt::AlignCenter);
	centralLayout->addWidget(statusLabel, 3, 0, 1, 3, Qt::AlignCenter);
	//

	// Pokud bu stacene tlacitko prevButton.
	connect(prevButton, SIGNAL(clicked()), this, SLOT(OnPrevButtonClick()));

	// Pokud bude stlaceno tlacitko nextButto.
	connect(nextButton, SIGNAL(clicked()), this, SLOT(OnNextButtonClick()));

	// Pokud bude stlaceno tlacitko startButton.
	connect(startButton, SIGNAL(clicked()), this, SLOT(OnStartButtonClick()));

	// Konec simulace.
	connect(this, SIGNAL(SimulationEnded()), this, SLOT(OnSimulationEnded()));
}

void MainDefWindow::OnPrevButtonClick()
{
	// Pokud neni den mimo vektor.
	if (currentDay > 0 && currentWeek >= 0) {
		// Ulozi se soucasny den.
		weeks[currentWeek]->getDay(currentDay % weeks[0]->getDays().size())->saveDay();
		// Pokud bude predchazet jiny tyden.
		if ((--currentDay % weeks[0]->getDays().size()) == 0 && currentWeek > 0) {
			--currentWeek;
		}
		// Dalis den.
		Day* temp = weeks[currentWeek]->getDay(currentDay % weeks[0]->getDays().size());
		// Nastavi se GUI predchazejiciho dne.
		stackedWidget->setCurrentIndex(currentDay);
		// Zmeni se progressbar.
		progress->setValue(progressValue -= progressIncrease);
		if (currentDay == 0) {
			// Dale nelze prejit na predchozi den, zablokuje se tlacitko.
			prevButton->setDisabled(true);
		}
		if (currentDay == weeks.size() * weeks[0]->getDays().size() - 2) {
			// Povoli lacitko dalis, pokud bylo nahodou zablokovano.
			nextButton->setDisabled(false);
		}
	}
}

void MainDefWindow::OnNextButtonClick()
{
	// Pokud neni den mimo vektor.
	if (currentWeek < weeks.size() && currentDay < (weeks.size() * weeks[0]->getDays().size() - 1)) {
		// Ulozi se soucasny den.
		weeks[currentWeek]->getDay(currentDay % weeks[0]->getDays().size())->saveDay();
		// Pokud bude nasledovat jiny tyden.
		if ((++currentDay % weeks[0]->getDays().size()) == 0) {
			++currentWeek;
		}
		// Nasledujici den.
		Day* temp = weeks[currentWeek]->getDay(currentDay % weeks[0]->getDays().size());
		// Nastavi se GUI nasledujiciho dne.
		stackedWidget->setCurrentIndex(currentDay);
		// Zmeni se progressbar.
		progress->setValue(progressValue += progressIncrease);
		if (currentDay > 0) {
			// Povoli se tlacitko zpet, pokud bylo nahodou zablokovano.
			prevButton->setDisabled(false);
		}
		if (currentDay == weeks.size() * weeks[0]->getDays().size() - 1) {
			// Zakblokuje tlacitko dalsi, dal nejde jit.
			nextButton->setDisabled(true);
		}
	}
}

void MainDefWindow::OnStartButtonClick()
{
	// Povoli se simulace.
	ready = 0;

	// Zablokuje se tlacitko, aby neslo spustit dalsi simulaci.
	startButton->setDisabled(true);

	// Vytvori se vlakno pro spusteni SmartCMGS simulaci.
	smartCGMSExecutionThread = std::make_unique<std::thread>(&MainDefWindow::SmartCGMS_Execution_Thread_Func, this, QStringToStdWString(fileNAmeEdit->text()));

	// Scope zamku - ceka se, az se simulace spusti.
	{
		std::unique_lock<std::mutex> lck(readyMtx);

		// Pocka se na SmartCGMS vlakno, az da vedet, ze bezi.
		size_t retries = 0;
		while (ready == 0 && (++retries) <= 5)
			readyCv.wait_for(lck, std::chrono::milliseconds(1000));

		// timeout
		if (ready == 0)
		{
			QMessageBox box(QMessageBox::Icon::Critical, "Error", "Could not start SmartCGMS core - timeout occurred!");
			box.exec();
			return;
		}

		// error
		if (ready < 0)
		{
			QMessageBox box(QMessageBox::Icon::Critical, "Error", "Could not start SmartCGMS core - and error occurred!");
			box.exec();
			return;
		}
	}

	// Label ukaze, ze simulace bezi.
	statusLabel->setText("Simulation in progress!");

	// Pocatecni casovy signal - 1. 1. 2020, 0:00:00.
	const double baseTime = Unix_Time_To_Rat_Time(1577833200);

	// Marker zacatku segmentu.
	Send_Time_Segment_Marker(true, baseTime);
	// Marker pro generatory signalu (aby se mohly synchronizovat).
	Send_Level(0.0, baseTime, scgms::signal_Null);

	int minutes, hours;
	double time;

	Week* thisWeek;
	Day* thisDay;
	DayEventValues* thisValue;

	// Pro vsechny tydny.
	for (int i = 0; i < weeks.size(); ++i) {
		thisWeek = weeks[i];
		// Pro vsechny dny tydne.
		for (int j = 0; j < thisWeek->getDays().size(); ++j) {
			thisDay = thisWeek->getDay(j);
			// Tyden je pospan sedmi dny.
			if (numOfDays == 7) {
				// Pro vsechny udalosti potravy dne.
				for (int k = 0; k < thisDay->getDayEvents().size() - 1; ++k) {
					thisValue = thisDay->getDayEvent(k);
					// Pokud je udalost oznacena.
					if (thisValue->getEventIsChecked()) {
						// Prida se randomizace.
						minutes = thisValue->getEventTimeMinutes() + randomize_time(thisValue->getEventJitter());
						hours = thisValue->getEventTimeHours();
						//Nastavi se cas pro SMartCGMS.
						time = baseTime + scgms::One_Hour * hours + scgms::One_Minute * minutes;
						// Posle se signal prjmu cukru.
						Send_Level(thisValue->getEventValue(), time, scgms::signal_Carb_Intake);
					}
				}
				thisValue = thisDay->getDayEvent(5);
				// Fyzicka aktivita.
				if (thisValue->getEventIsChecked()) {
					// Prida se randomizace.
					minutes = randomize_time(thisValue->getEventJitter());
					hours = thisValue->getEventTimeHours();
					//Nastavi se cas pro SMartCGMS.
					time = baseTime + scgms::One_Hour * hours + scgms::One_Minute * minutes;
					// Posle se signal prjmu aktivity.
					Send_Level(thisValue->getEventValue(), time, scgms::signal_Physical_Activity);
				}
			}
			// Tyden je popsaj jednim dnem.
			else {
				// Vse probehne sedmkrat.
				for (int x = 0; x < 7; ++x) {
					// Pro vsechny udalosti potravy dne.
					for (int k = 0; k < thisDay->getDayEvents().size() - 1; ++k) {
						thisValue = thisDay->getDayEvent(k);
						// Pokud je udalost oznacena.
						if (thisValue->getEventIsChecked()) {
							// Prida se randomizace.
							minutes = randomize_time(thisValue->getEventJitter());
							hours = thisValue->getEventTimeHours();
							//Nastavi se cas pro SMartCGMS.
							time = baseTime + scgms::One_Hour * hours + scgms::One_Minute * minutes;
							// Posle se signal prjmu cukru.
							Send_Level(thisValue->getEventValue(), time, scgms::signal_Carb_Intake);
						}
					}
					thisValue = thisDay->getDayEvent(5);
					// Pokud je udalost oznacena.
					if (thisValue->getEventIsChecked()) {
						// Prida se randomizace.
						minutes = randomize_time(thisValue->getEventJitter());
						hours = thisValue->getEventTimeHours();
						//Nastavi se cas pro SMartCGMS.
						time = baseTime + scgms::One_Hour * hours + scgms::One_Minute * minutes;
						// Posle se signal prjmu aktivity.
						Send_Level(thisValue->getEventValue(), time, scgms::signal_Physical_Activity);
					}
				}
			}
		}
	}

	// Marker pro generatory.
	Send_Level(0.0, baseTime + scgms::One_Hour * 24.0, scgms::signal_Null);
	// Marker konce segmentu.
	Send_Time_Segment_Marker(false, baseTime + scgms::One_Hour * 24.0);

	// Korektne ukonci simulaci.
	Send_Shut_Down(baseTime + scgms::One_Hour * 24.0);
}

// spojka pro hledani filtru - jen zavola metodu okna
HRESULT IfaceCalling On_Filter_Created_Bridge_(scgms::IFilter* filter, const void* data)
{
	MainDefWindow* w = const_cast<MainDefWindow*>(reinterpret_cast<const MainDefWindow*>(data));

	return S_OK;
}

void MainDefWindow::SmartCGMS_Execution_Thread_Func(const std::wstring& fileName)
{
	scgms::SPersistent_Filter_Chain_Configuration configuration;
	refcnt::Swstr_list errors;

	// Pomocna funkce - tou budeme notifikovat vnejsi kod, kdyz budeme ready nebo kdyz narazime na chybu.
	auto notifyReady = [this](int readyStatus) {
		std::unique_lock<std::mutex> lck(readyMtx);
		ready = readyStatus;
		readyCv.notify_all();
	};
	// Pomocna funkce - vypise na konzoli vsechny chyby, pokud nejake jsou.
	auto printErrorsIfAny = [errors]() {
		errors.for_each([](auto str) {
			std::wcerr << str << std::endl;
			});
	};

	// Nacte konfiguraci ze souboru.
	HRESULT rc{};

	// Nacte nkofiguraci z pameti.
	rc = configuration->Load_From_Memory(BaseConfig::Generator_Config.c_str(), BaseConfig::Generator_Config.length(), errors.get());

	// Nastavi promennou nazvu vytvoreneho souboru.
	configuration->Set_Variable(BaseConfig::Generator_Config_Log_File_Name_Var, fileName.c_str());

	// Vypise chyby, pokud jsou.
	printErrorsIfAny();

	// Neco se nezdarilo, vypise se hlaska a ukonci se metoda.
	if (!Succeeded(rc)) {
		std::wcerr << L"Cannot load the configuration; Error code: " << rc << std::endl;
		notifyReady(-1);
		return;
	}

	// Neco se pokazilo, ale ne tak, aby nas to trapilo nejak vic.
	if (rc == S_FALSE) {
		std::wcerr << L"Warning: some filters were not fully loaded!" << std::endl;
	}

	// Spousti se simulace.
	filterExecutor = scgms::SFilter_Executor(configuration, &On_Filter_Created_Bridge_, this, errors, dynamic_cast<scgms::IFilter*>(this));
	printErrorsIfAny();

	// NAstal problem, vypise se hlaska a ukonci se metoda.
	if (!filterExecutor) {
		std::wcerr << L"Could not execute the filters!" << std::endl;
		notifyReady(-1);
		return;
	}

	// Notifikujeme, ze vsechno bezi.
	notifyReady(1);
	// Zabrzdi se vlakno, dokud se simulace neukonci.
	filterExecutor->Terminate(TRUE);
	// Jakmile simulace skonci, notifikujeme GUI, ze je hotovo.
	emit SimulationEnded();
}

void MainDefWindow::OnSimulationEnded()
{
	// Nastavi se label, ze simulace skoncila.
	statusLabel->setText("Simulation finished!");

	// Pro jistotu se posle shutdown zprava, pokud treba v konfiguraci chybi.
	if (filterExecutor)
		Send_Shut_Down();

	int joinable = 0;

	// Vlakno se nastavi na join stav.
	if (smartCGMSExecutionThread && smartCGMSExecutionThread->joinable()) {
		smartCGMSExecutionThread->join();
	}

	// Zahodi se odkazy na SCGMS exekutor a vlakno.
	filterExecutor.reset();
	smartCGMSExecutionThread.reset();

	// Povoli se dalsi simulace.
	startButton->setDisabled(false);
}

bool MainDefWindow::Send_Shut_Down(double time)
{
	if (!filterExecutor)
		return false;

	// Vtyvori se UDevice_Event.
	scgms::UDevice_Event sd{ scgms::NDevice_Event_Code::Shut_Down };
	// Nepodarilo se ho vytvorit, nastavi se device_time.
	if (!std::isnan(time))
		sd.device_time() = time;

	return Succeeded(filterExecutor.Execute(std::move(sd)));
}

bool MainDefWindow::Send_Time_Segment_Marker(bool start, double time)
{
	// Nastavi se promenne UDevice_event.
	scgms::UDevice_Event evt{ start ? scgms::NDevice_Event_Code::Time_Segment_Start : scgms::NDevice_Event_Code::Time_Segment_Stop };
	evt.device_time() = time;
	evt.segment_id() = 1;
	evt.device_id() = BaseConfig::Generator_GUID;
	return Succeeded(filterExecutor.Execute(std::move(evt)));
}

bool MainDefWindow::Send_Level(double value, double time, const GUID& signal_id)
{
	// Nastavi se promenne UDevice_event.
	scgms::UDevice_Event evt{ scgms::NDevice_Event_Code::Level };
	evt.device_time() = time;
	evt.signal_id() = signal_id;
	evt.level() = value;
	evt.segment_id() = 1;
	evt.device_id() = BaseConfig::Generator_GUID;
	return Succeeded(filterExecutor.Execute(std::move(evt)));
}

int MainDefWindow::randomize_time(int jitter_value) {
	// Neni potreba vytvaret nahodne cislo.
	if (jitter_value == 0) {
		return 0;
	}

	// Nahodne se vybere znamenko.
	int mark = rand() % 2;

	// Generator nahodnych cisel.
	std::mt19937 rng(dev());
	// Nastavi se rozmezi.
	std::uniform_int_distribution<std::mt19937::result_type> dist6(0, jitter_value - 1);

	// Vygeneruje se hodnota.
	int rand_value = dist6(rng);

	// Vrati se hodnota i se znamenkem.
	return (mark == 0) ? rand_value : -1 * rand_value;
}

void MainDefWindow::setAgeValue(int ageValue)
{
	this->ageValue = ageValue;
}

void MainDefWindow::setDiabetValue(bool diabetValue)
{
	this->diabetValue = diabetValue;
}

void MainDefWindow::setNumOfDays(int numOfDays)
{
	this->numOfDays = numOfDays;
}

void MainDefWindow::setNumOfWeeks(int numOfWeeks)
{
	this->numOfWeeks = numOfWeeks;
}
