
#include "hausdorff_distance_man.h"

#include <cmath>

Hausdorff_Distance_Manhattan::Hausdorff_Distance_Manhattan(const scgms::TMetric_Parameters& params) : mParameters{ params } {

}

Hausdorff_Distance_Manhattan::~Hausdorff_Distance_Manhattan() {
	// Vycisti se vsechny vektory.
	Reference.clear();
	Calculated.clear();
	Times.clear();
	Minims.clear();
}

double Hausdorff_Distance_Manhattan::Manhattan_Distance(double reference, double calculated, double time_i, double time_j)
{
	double dist, diffY, diffX;

	// Rozdil osy X.
	diffX = time_i - time_j;

	// Rozdil osy Y.
	diffY = reference - calculated;

	// Soucet obsolutnich hodnot vzdalenosti.
	dist = std::abs(diffX) + std::abs(diffY);

	return dist;
}

void Hausdorff_Distance_Manhattan::FindMinims()
{
	double temp_min, current;

	// Permutace vsech  hodnot vzhledem k prvni krivce, z nich se vyberou ty nejvetsi.
	for (int i = 0; i < Reference.size(); ++i) {
		temp_min = DBL_MAX;
		for (int j = 0; j < Calculated.size(); ++j) {
			// Vypocet vzdalenosti mezi dvema body.
			current = Manhattan_Distance(Reference[i], Calculated[j], Times[i], Times[j]);
			if (current < temp_min) {
				temp_min = current;
			}
		}
		Minims.push_back(temp_min);
	}

	// Permutace vsech  hodnot vzhledem k druhe krivce, z nich se vyberou ty nejvetsi.
	for (int i = 0; i < Calculated.size(); ++i) {
		temp_min = DBL_MAX;
		for (int j = 0; j < Reference.size(); ++j) {
			// Vypocet vzdalenosti mezi dvema body.
			current = Manhattan_Distance(Calculated[i], Reference[j], Times[i], Times[j]);
			if (current < temp_min) {
				temp_min = current;
			}
		}
		//Nejmensi hodnota se prida do vektoru.
		Minims.push_back(temp_min);
	}

	// Vycisti se vektroy hodnot.
	Reference.clear();
	Calculated.clear();
	Times.clear();
}

HRESULT IfaceCalling Hausdorff_Distance_Manhattan::Accumulate(const double* times, const double* reference, const double* calculated, const size_t count)
{
	// Predaji se ziskane hodnoty pro nasledny vypocet.
	for (int i = 0; i < count; ++i) {
		Reference.push_back(reference[i]);
		Calculated.push_back(calculated[i]);
		Times.push_back(times[i]);
	}

	return S_OK;
}

HRESULT IfaceCalling Hausdorff_Distance_Manhattan::Reset()
{
	// Vycisti se vsechny vektory.
	Minims.clear();
	Reference.clear();
	Calculated.clear();
	Times.clear();

	return S_OK;
}

HRESULT IfaceCalling Hausdorff_Distance_Manhattan::Calculate(double* metric, size_t* levels_accumulated, size_t levels_required)
{
	// Ziskal se pozadovany pocet hdnot?
	*levels_accumulated = Reference.size();
	if (*levels_accumulated < std::max(static_cast<size_t>(1), levels_required)) {
		return S_FALSE;
	}

	// Najdou se vsechny minima.
	FindMinims();

	// Z minim se vybere nejvetsi hodnota.
	for (int i = 0; i < *levels_accumulated; ++i) {
		if (Minims[i] > metricValue) {
			metricValue = Minims[i];
		}
	}

	// Preda se vysledna hodnota.
	*metric = metricValue;

	return S_OK;
}

HRESULT IfaceCalling Hausdorff_Distance_Manhattan::Get_Parameters(scgms::TMetric_Parameters* parameters) {

	// Zkopiruji se parametry.
	memcpy(parameters, &mParameters, sizeof(mParameters));

	return S_OK;
}
