
#pragma once

#include <rtl/FilterLib.h>
#include <rtl/referencedImpl.h>
#include <rtl/SolverLib.h>
#include <rtl/UILib.h>

#include <limits>

#pragma warning( push )
#pragma warning( disable : 4250 ) // C4250 - 'class1' : inherits 'class2::member' via dominance

/* Definice tridy implementujici Hausdorffovu vzdalenostmodifikovana Manahttanskou metrikou. */
class Hausdorff_Distance_Manhattan : public virtual scgms::IMetric, public virtual refcnt::CReferenced {

protected:

	/* Parametr pro SmartCMGS */
	const scgms::TMetric_Parameters mParameters;

	/* Vektor refenrcnich hodnot. */
	std::vector<double> Reference;

	/* Vektor vypocitanych hodnot. */
	std::vector<double> Calculated;

	/* Vektor casu. */
	std::vector<double> Times;

	/* Vektor minim dane iterace. */
	std::vector<double> Minims;
public:
	/* Vytvori novou instanci objektu. */
	Hausdorff_Distance_Manhattan(const scgms::TMetric_Parameters& params);
	virtual ~Hausdorff_Distance_Manhattan();

	/* Vypocte mahattanskou metriku mezi dvema body. */
	double Manhattan_Distance(double reference, double calculated, double time_i, double time_j);

	/* Nalezne minima pro kazdou iteraci. */
	void FindMinims();

	/* Preradi zsikane hodnoty. */
	virtual HRESULT IfaceCalling Accumulate(const double* times, const double* reference, const double* calculated, const size_t count) final;
	
	/* Vycisti vektory hodnot. */
	virtual HRESULT IfaceCalling Reset() final;
	
	/* Vypocita hodnotu metriky. */
	virtual HRESULT IfaceCalling Calculate(double* metric, size_t* levels_accumulated, size_t levels_required) final;
	
	/* Vrati parametr. */
	virtual HRESULT IfaceCalling Get_Parameters(scgms::TMetric_Parameters* parameters) final;

private:

	/* Hodnota nalezeneho maxima (hodnota metriky). */
	double metricValue = 0;
};

#pragma warning( pop )